from herre.config import HerreConfig
from herre.herre import Herre


def build_authorize_url(herre: Herre):
    return f"{herre.base_url}/{herre.auth_path}/" if herre.append_trailing_slash else f"{herre.base_url}/{herre.auth_path}"

def build_token_url(herre: Herre):
    return f"{herre.base_url}/{herre.token_path}/" if herre.append_trailing_slash else f"{herre.base_url}/{herre.token_path}"

def build_refresh_url(herre: Herre):
    return f"{herre.base_url}/{herre.refresh_path}/" if herre.append_trailing_slash else f"{herre.base_url}/{herre.refresh_path}"



from typing import List, Optional
from pydantic import BaseModel


class App(BaseModel):
    clientID: Optional[str]
    name: Optional[str]


class User(BaseModel):
    username: Optional[str]
    email: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]


class HerreState(BaseModel):
    user: Optional[User]
    scopes: List[str]
    client_id: str
    access_token: str
    refresh_token: Optional[str]



from enum import Enum
from urllib.parse import quote
from aiohttp import web
import asyncio
import webbrowser
import uuid


class RedirectError(Exception):
    pass


def wrapped_qs_future(future):
    async def web_token_response(request):
        future.set_result(request.path_qs)
        return web.Response(text="You can close me now !")

    return web_token_response


async def wait_for_redirect(
    starturl,
    redirect_host="localhost",
    redirect_port=6767,
    path="/",
    timeout=400,
    print_function=False,
    handle_signals=False,
):

    webbrowser.open(starturl)

    token_future = asyncio.get_event_loop().create_future()

    app = web.Application()

    app.router.add_get(path, wrapped_qs_future(token_future))

    webserver_future = asyncio.wait_for(
        web._run_app(
            app,
            host=redirect_host,
            port=redirect_port,
            print=print_function,
            handle_signals=handle_signals,
        ),
        timeout,
    )
    done, pending = await asyncio.wait(
        [token_future, webserver_future], return_when=asyncio.FIRST_COMPLETED
    )

    for tf in done:
        if tf == token_future:
            redirect_qs = tf.result()
        else:
            redirect_qs = None

    for task in pending:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

    if not redirect_qs:
        raise RedirectError("No redirect Queryset provided")
    return redirect_qs



from fakts.fakts import Fakts, get_current_fakts
from herre.herre import Herre
from herre.fakts.config import HerreConfig
from herre.fakts.registry import GrantRegistry, get_current_grant_registry


class FaktsHerre(Herre):

    def __init__(self, *args, fakts: Fakts = None, grant_registry: GrantRegistry = None, fakts_key="herre", **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._grant_registry = grant_registry or get_current_grant_registry()
        self._fakts = fakts or get_current_fakts()
        self.config = None
        self._fakts_key = fakts_key

    def configure(self, config: HerreConfig) -> None:
        self.base_url = config.base_url
        self.client_id = self.client_id or config.client_id
        self.client_secret = self.client_secret or config.client_secret

        self.grant = self.grant or self._grant_registry.get_grant_for_type(
            config.authorization_grant_type
        )(**config.grant_kwargs)
        print(self.grant)

        self.configured = True
        self.config = config

    async def alogin(self, **kwargs):

        if not self.config:
            self.fakts = await self._fakts.aget(self._fakts_key)
            self.configure(HerreConfig(**self.fakts))

        return await super().alogin(**kwargs)



from abc import abstractmethod
from typing import Any, List
from herre.config import HerreConfig
from abc import ABC
import os
import logging
import aiohttp

logger = logging.getLogger(__name__)


class GrantException(Exception):
    pass


class RetryException(GrantException):
    pass


class BaseGrant(ABC):
    herre: Any
    client_id: str
    client_secret: str
    scopes: List[str]

    @abstractmethod
    async def afetch_token(self, herre, **kwargs):
        raise NotImplementedError("Impelment afetch_token")



from enum import Enum
from typing import Any, Dict, List, Optional
from pydantic import BaseModel


class GrantType(str, Enum):
    IMPLICIT = "IMPLICIT"
    PASSWORD = "PASSWORD"
    CLIENT_CREDENTIALS = "CLIENT_CREDENTIALS"
    AUTHORIZATION_CODE = "AUTHORIZATION_CODE"
    AUTHORIZATION_CODE_SERVER = "AUTHORIZATION_CODE_SERVER"


class HerreConfig(BaseModel):
    base_url: str
    client_id: str
    client_secret: str
    authorization_grant_type: GrantType
    grant_kwargs: Dict[str, Any] = {}
    scopes: List[str]
    redirect_uri: Optional[str]
    jupyter_sync: bool = False
    username: Optional[str]
    password: Optional[str]
    timeout: int = 500
    no_temp: bool = False



from herre.grants.refreshable import RefreshableGrant
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from herre.grants.session import OAuth2Session

from herre.grants.registry import register_grant
from herre.config import GrantType
from herre.grants.utils import build_token_url
from herre.herre import Herre


class BackendGrant(RefreshableGrant):
    refreshable = True

    async def afetch_token(self, herre: Herre, **kwargs):
        auth_client = BackendApplicationClient(
            client_id=herre.client_id, scope=" ".join(herre.requested_scopes)
        )

        async with OAuth2Session(client=auth_client, scope=herre.scope) as session:
            print(build_token_url(herre))

            token = await session.fetch_token(
                token_url=build_token_url(herre),
                client_id=herre.client_id,
                client_secret=herre.client_secret,
                verify=True,
            )

        return token



from typing import Dict, Type
from herre.grants.base import BaseGrant
from herre.config import GrantType
import logging


logger = logging.getLogger(__name__)


class GrantRegistry:

    def __init__(self, register_defaults= True) -> None:
        self.registeredGrants: Dict[str, BaseGrant] = {}
        if register_defaults:
            self.register_defaults()

    def register_defaults(self):
        from herre.grants.backend.app import BackendGrant
        from herre.grants.code_server.app import AuthorizationCodeServerGrant

        self.register_grant(GrantType.AUTHORIZATION_CODE, AuthorizationCodeServerGrant)
        self.register_grant(GrantType.CLIENT_CREDENTIALS, BackendGrant)


    def register_grant(self, type: GrantType, grant: Type[BaseGrant]):
        self.registeredGrants[type] = grant

    def get_grant_for_type(self, type):
        return self.registeredGrants[type]


def register_grant(type: GrantType):
    def rea_decorator(grant):
        assert hasattr(grant, "afetchtoken"), "A grant must specify a afetchtoken method"
        logger.info(f"Registering Grant {grant} for {type}")
        get_current_grant_registry().register_grant(type, grant)
        return grant

    return rea_decorator


GRANT_REGISTRY = None


def get_current_grant_registry(with_defaults=True):
    global GRANT_REGISTRY
    if not GRANT_REGISTRY:
        GRANT_REGISTRY = GrantRegistry()

    return GRANT_REGISTRY



import aiohttp
from herre.errors import NoHerreFound
from herre.grants.base import BaseGrant
import os
import logging
from herre.types import HerreState, User
import shelve
import contextvars
import os
from koil.loop import koil


current_herre = contextvars.ContextVar("current_herre", default=None)
GLOBAL_HERRE = None


logger = logging.getLogger(__name__)


def set_current_herre(herre, set_global=True):
    global GLOBAL_HERRE
    current_herre.set(herre)
    if set_global:
        GLOBAL_HERRE = herre


def set_global_herre(herre):
    global GLOBAL_HERRE
    GLOBAL_HERRE = herre


def get_current_herre(allow_global=True):
    global GLOBAL_HERRE
    herre = current_herre.get()

    if not herre:
        if not allow_global:
            raise NoHerreFound(
                "No current herre found and global herre are not allowed"
            )
        if not GLOBAL_HERRE:
            if os.getenv("HERRE_ALLOW_FAKTS_GLOBAL", "True") == "True":
                try:
                    from herre.fakts import FaktsHerre

                    GLOBAL_HERRE = FaktsHerre()
                    return GLOBAL_HERRE
                except ImportError as e:
                    raise NoHerreFound("Error creating Fakts Herre") from e
            else:
                raise NoHerreFound(
                    "No current herre found and and no global herre found"
                )

        return GLOBAL_HERRE

    return herre


class Herre:
    def __init__(
        self,
        *args,
        grant: BaseGrant = None,
        base_url: str = "",
        client_id: str = "",
        client_secret: str = "",
        scopes=["introspection"],
        token_path="token",
        authorize_path="authorize",
        refresh_path="token",
        append_trailing_slash=True,
        token_file="token.temp",
        userinfo_path="userinfo",
        max_retries=5,
        allow_insecure=False,
        no_temp=False,
        register_global=False,
        **kwargs,
    ) -> None:
        """Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """

        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if allow_insecure else "1"

        self.max_retries = max_retries
        self.base_url = base_url  # is defered
        self.client_id = client_id
        self.client_secret = client_secret
        self.requested_scopes = scopes
        self.scope = " ".join(self.requested_scopes)
        self.grant = grant
        self.base_url = base_url
        self.auth_path = authorize_path
        self.token_path = token_path
        self.refresh_path = refresh_path
        self.userinfo_path = userinfo_path
        self.append_trailing_slash = append_trailing_slash
        self.token_file = token_file
        self.no_temp = no_temp

        self.state: HerreState = None

        super().__init__(*args, **kwargs)
        if register_global:
            set_global_herre(self)

    async def aget_token(self):
        if not self.state or not self.state.access_token:
            await self.alogin()

        return self.state.access_token

    async def alogin(self, force_relogin=False, retry=0, **kwargs) -> HerreState:

        if retry > self.max_retries:
            raise Exception("Exceeded Login Retries")

        if not force_relogin and not self.no_temp:
            try:
                with shelve.open(self.token_file) as cfg:
                    client_id = cfg["client_id"]
                    if client_id == self.grant.client_id:
                        self.state = HerreState(**cfg["state"])
                    else:
                        logger.info("Omitting old token")

            except Exception:
                pass

        if not self.state:
            token_dict = await self.grant.afetch_token(self, **kwargs)
            logger.info(f"Grant fetched token {token_dict}")
            self.state = HerreState(
                **token_dict, client_id=self.client_id, scopes=self.requested_scopes
            )

        try:
            user_info_endpoint = build_userinfo_url(self)
            async with aiohttp.ClientSession(
                headers={"Authorization": f"Bearer {self.state.access_token}"}
            ) as session:
                async with session.get(user_info_endpoint) as resp:
                    user_json = await resp.json()
                    if "detail" in user_json:
                        raise Exception(user_json["detail"])

                    try:
                        self.state.user = User(**user_json)
                    except:
                        self.state.user = None

                    if not self.no_temp:
                        with shelve.open(self.token_file) as cfg:
                            cfg["client_id"] = self.state.client_id
                            cfg["state"] = self.state.dict()

        except Exception as e:
            logger.error(f"Token Invalid {e}")
            self.state = None
            await self.alogin(force_relogin=True, retry=retry + 1, **kwargs)

        return self.state

    async def alogout(self):
        if self.grant:
            return await self.grant.alogout()

        try:
            with shelve.open(self.token_file) as cfg:
                cfg["state"] = None
                cfg["client_id"] = None
        except KeyError:
            pass

        self.state = None

    async def arefresh(self):
        return await self.grant.alogin()

    def login(self, **kwargs):
        return koil(self.alogin(), **kwargs)

    def logout(self, **kwargs):
        return koil(self.alogout(), **kwargs)

    def refresh(self):
        return koil(self.arefresh())

    @property
    def logged_in(self):
        return self.state is not None

    @property
    def user(self):
        assert self.state, "We are not yet logged in"
        assert self.state.user, "Login is not associated with a user"
        return self.state.user

    @property
    def scopes(self):
        assert self.state, "We are not yet logged in"
        assert self.state.scopes, "Login is not associated with a user"
        return self.state.scopes

    @property
    def headers(self):
        assert self.state, "We are not yet logged in"
        return {"Authorization": f"Bearer {self.state.access_token}"}


def build_userinfo_url(herre: Herre):
    return (
        f"{herre.base_url}/{herre.userinfo_path}/"
        if herre.append_trailing_slash
        else f"{herre.base_url}/{herre.userinfo_path}"
    )



from herre.config import GrantType
from herre.grants.refreshable import RefreshableGrant
from oauthlib.oauth2 import WebApplicationClient
from herre.grants.session import OAuth2Session
import webbrowser
from aiohttp import web
import asyncio
import logging
from herre.grants.registry import register_grant
from herre.grants.utils import build_authorize_url, build_token_url
from herre.herre import Herre

logger = logging.getLogger(__name__)
REDIRECT_PORT = 6767


def wrapped_future(future):
    async def web_token_response(request):
        future.set_result(request.path_qs)
        return web.Response(text="You can close me now !")

    return web_token_response


async def wait_for_server(app, host="localhost", port="6767", timeout=1):
    try:
        await asyncio.wait_for(
            web._run_app(
                app,
                host="localhost",
                port=6767,
            ),
            timeout,
        )
    except asyncio.TimeoutError:
        return "no token"


class AuthorizationCodeServerGrant(RefreshableGrant):
    refreshable = True

    def __init__(
        self,
        redirect_port=REDIRECT_PORT,
        redirect_timeout=40,
    ) -> None:
        self.redirect_port = redirect_port
        self.redirect_timeout = redirect_timeout
        self.redirect_uri = f"http://localhost:{redirect_port}/"

    async def afetch_token(self, herre: Herre):

        web_app_client = WebApplicationClient(
            herre.client_id, scope=herre.scope
        )

        # Create an OAuth2 session for the OSF
        async with OAuth2Session(
            herre.client_id,
            web_app_client,
            scope=herre.scope,
            redirect_uri=self.redirect_uri,
        ) as session:

            auth_url, state = session.authorization_url(build_authorize_url(herre))
            print(auth_url)
            webbrowser.open(auth_url)

            token_future = asyncio.get_event_loop().create_future()

            app = web.Application()
            app.router.add_get("/", wrapped_future(token_future))
            webserver_future = asyncio.wait_for(
                web._run_app(
                    app,
                    host="localhost",
                    port=self.redirect_port,
                    print=lambda x: logger.info(x),
                    handle_signals=False,
                ),
                self.redirect_timeout,
            )
            done, pending = await asyncio.wait(
                [token_future, webserver_future], return_when=asyncio.FIRST_COMPLETED
            )

            for tf in done:
                if tf == token_future:
                    path = tf.result()
                else:
                    path = None

            for task in pending:
                task.cancel()

                try:
                    await task
                except asyncio.CancelledError:
                    pass

            if path:
                return await session.fetch_token(
                    build_token_url(herre),
                    client_secret=herre.client_secret,
                    authorization_response=path,
                    state=state,
                )

        return None



from herre import Herre
from qtpy import QtCore, QtWidgets


class QtHerre(Herre, QtWidgets.QWidget):
    login_signal = QtCore.Signal()
    logout_signal = QtCore.Signal()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    async def alogin(self, **kwargs):
        nana = await super().alogin(**kwargs)
        self.login_signal.emit()
        return nana

    async def alogout(self, **kwargs):
        nana = await super().alogout(**kwargs)
        self.logout_signal.emit()
        return nana



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['herre',
 'herre.fakts',
 'herre.grants',
 'herre.grants.backend',
 'herre.grants.code_server']

package_data = \
{'': ['*']}

install_requires = \
['QtPy>=1.11.1,<2.0.0',
 'aiohttp>=3.7.4,<4.0.0',
 'oauthlib>=3.1.1,<4.0.0',
 'rich>=10.9.0,<11.0.0']

extras_require = \
{':extra == "fakts"': ['fakts==0.1.40']}

setup_kwargs = {
    'name': 'herre',
    'version': '0.1.63',
    'description': 'The governing package for modules of the arnheim framework, it provides core features like wards, the event loop and authentication with an oauth provider',
    'long_description': '# herre\n\n[![codecov](https://codecov.io/gh/jhnnsrs/herre/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/herre)\n[![PyPI version](https://badge.fury.io/py/herre.svg)](https://pypi.org/project/herre/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/herre.svg)](https://pypi.python.org/pypi/herre/)\n[![PyPI status](https://img.shields.io/pypi/status/herre.svg)](https://pypi.python.org/pypi/herre/)\n[![PyPI download month](https://img.shields.io/pypi/dm/herre.svg)](https://pypi.python.org/pypi/herre/)\n\n#### DEVELOPMENT\n\n## Idea\n\nherre is an (asynchronous) oauth2/openid client, that provides sensible defaults for the python\necosystem\n\n## Prerequisites\n\nherre needs a oauth2 server to connect to\n\n## Supports\n\n- Authorization Code Flow (PKCE)\n- Client-Credentials Flow\n\n## Usage\n\nIn order to initialize the Client you need to connect it as a Valid Application with your Arnheim Instance\n\n```python\n\nclient = Herre(\n    grant=AuthorizationCode()\n    host="p-tnagerl-lab1",\n    port=8000,\n    client_id="$YOUR_CLIENT_ID",\n    client_secret="$YOUR_CLIENT_SECRET",\n    name="karl",\n)\n\nclient.login()\n\n```\n\n## Intergration with Qt\n\nherre fully supports qt-based applications (both PySide2 and PyQt5) and provides a convenient helper class \'QtHerre\'\nas well as a included windowed Authoriation Code Flow (needs pyqtwebengine as additional dependency) as well as browser based logins\n\n```python\nclass MainWindow(QtWidget)\n\n    def __init__(self, *args, **kwargs):\n        self.herre = QtHerre(\n          grant=QtWindowAuthorizationCode()\n        )\n\n```\n\n## Build with\n\n- [koil](https://github.com/jhnnsrs/koil) - for pyqt event loop handling\n- [oauthlib](https://github.com/oauthlib/oauthlib) - for oauth2 compliance\n- [aiohttp](https://github.com/aio-libs/aiohttp) - transport layer (especially redirect server)\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
import aiohttp
from herre.herre import Herre


class RefreshableGrant:
    type = None

    @abstractmethod
    async def afetch_token(self, herre: Herre):
        raise NotImplementedError()

    async def arefresh(self, herre: Herre):
        assert "refresh_token" in self.token, "Token had not refresh-token attached"
        async with aiohttp.ClientSession() as session:
            async with session.post(
                herre.refresh_url,
                data={
                    "grant_type": "refresh_token",
                    "refresh_token": self.token["refresh_token"],
                    "client_id": herre.client_id,
                    "client_secret": herre.client_secret,
                },
            ) as resp:
                self.token = await resp.json()
                assert (
                    "access_token" in self.token
                ), "Returned refreshed token does not have an access_token"
                if "refresh_token" in self.token:
                    self.can_refresh = True

    
